@@ -5,7 +5,8 @@
 //! 3. Lua utitlites, implemented in Tarantool
 
 #![allow(non_camel_case_types)]
-use std::os::raw::{c_double, c_int, c_schar, c_void};
+#![allow(clippy::missing_safety_doc)]
+use std::os::raw::{c_double, c_int, c_char, c_void};
 use std::ptr::null_mut;
 
 /// Lua provides a registry, a pre-defined table that can be used by any C code
@@ -23,7 +24,7 @@ pub const LUA_REGISTRYINDEX: c_int = -10000;
 pub const LUA_ENVIRONINDEX: c_int = -10001;
 pub const LUA_GLOBALSINDEX: c_int = -10002;
 
-pub unsafe fn is_relative_index(index: c_int) -> bool {
+pub fn is_relative_index(index: c_int) -> bool {
     index < 0 && index > LUA_REGISTRYINDEX
 }
 
@@ -62,7 +63,7 @@ pub struct lua_State {
 
 #[repr(C)]
 pub struct luaL_Reg {
-    pub name: *const c_schar,
+    pub name: *const c_char,
     pub func: lua_CFunction,
 }
 
@@ -136,6 +137,15 @@ extern "C" {
     // Lua C API functions.
     pub fn lua_newstate(f: lua_Alloc, ud: *mut libc::c_void) -> *mut lua_State;
     pub fn lua_close(l: *mut lua_State);
+
+    /// Creates a new thread, pushes it on the stack, and returns a pointer to a
+    /// `lua_State` that represents this new thread. The new state returned by
+    /// this function shares with the original state all global objects (such as
+    /// tables), but has an independent execution stack.
+    /// *[-0, +1, m]*
+    ///
+    /// There is no explicit function to close or to destroy a thread. Threads
+    /// are subject to garbage collection, like any Lua object.
     pub fn lua_newthread(l: *mut lua_State) -> *mut lua_State;
 
     pub fn lua_atpanic(l: *mut lua_State, panicf: lua_CFunction) -> lua_CFunction;
@@ -157,8 +167,14 @@ extern "C" {
     /// string cannot contain embedded zeros; it is assumed to end at the first
     /// zero.
     /// *[-0, +1, m]*
-    pub fn lua_pushstring(l: *mut lua_State, s: *const c_schar) -> *const c_schar;
+    pub fn lua_pushstring(l: *mut lua_State, s: *const c_char) -> *const c_char;
+
+    /// Pushes a number with value `n` onto the stack.
+    /// *[-0, +1, -]*
     pub fn lua_pushinteger(l: *mut lua_State, n: isize);
+
+    /// Pushes a number with value `n` onto the stack.
+    /// *[-0, +1, -]*
     pub fn lua_pushnumber(l: *mut lua_State, n: c_double);
 
     /// Pushes a new C closure onto the stack.
@@ -199,7 +215,7 @@ extern "C" {
     /// garbage collection, there is no guarantee that the pointer returned by
     /// `lua_tolstring` will be valid after the corresponding value is removed
     /// from the stack.
-    pub fn lua_tolstring(l: *mut lua_State, index: c_int, len: *mut usize) -> *const c_schar;
+    pub fn lua_tolstring(l: *mut lua_State, index: c_int, len: *mut usize) -> *const c_char;
 
     /// If the value at the given acceptable `index` is a full userdata, returns
     /// its block address. If the value is a light userdata, returns its
@@ -213,14 +229,20 @@ extern "C" {
     ///
     /// This function pops the value from the stack. As in Lua, this function
     /// may trigger a metamethod for the "newindex" event
-    pub fn lua_setfield(l: *mut lua_State, index: c_int, k: *const c_schar);
+    pub fn lua_setfield(l: *mut lua_State, index: c_int, k: *const c_char);
 
     /// Pushes onto the stack the value `t[k]`, where `t` is the value at the
     /// given valid `index`. As in Lua, this function may trigger a metamethod
     /// for the "index" event
     /// *[-0, +1, e]*
-    pub fn lua_getfield(l: *mut lua_State, index: c_int, k: *const c_schar);
+    pub fn lua_getfield(l: *mut lua_State, index: c_int, k: *const c_char);
 
+    /// Creates a new empty table and pushes it onto the stack. The new table
+    /// has space pre-allocated for `narr` array elements and `nrec` non-array
+    /// elements. This pre-allocation is useful when you know exactly how many
+    /// elements the table will have. Otherwise you can use the function
+    /// [`lua_newtable`].
+    /// *[-0, +1, m]*
     pub fn lua_createtable(l: *mut lua_State, narr: c_int, nrec: c_int);
 
     /// This function allocates a new block of memory with the given size,
@@ -238,6 +260,16 @@ extern "C" {
     /// collected again then Lua frees its corresponding memory.
     pub fn lua_newuserdata(l: *mut lua_State, sz: libc::size_t) -> *mut libc::c_void;
 
+    /// Pushes a light userdata onto the stack.
+    /// *[-0, +1, -]*
+    ///
+    /// Userdata represent C values in Lua. A light userdata represents a
+    /// pointer. It is a value (like a number): you do not create it, it has no
+    /// individual metatable, and it is not collected (as it was never created).
+    /// A light userdata is equal to "any" light userdata with the same C
+    /// address.
+    pub fn lua_pushlightuserdata(l: *mut lua_State, ud: *mut c_void);
+
     /// Pushes onto the stack the value `t[k]`, where `t` is the value at the
     /// given valid `index` and `k` is the value at the top of the stack.
     /// *[-1, +1, e]*
@@ -292,7 +324,7 @@ extern "C" {
     /// Returns the name of the type encoded by the value `tp`, which must be
     /// one the values returned by [`lua_type`].
     /// *[-0, +0, -]*
-    pub fn lua_typename(state: *mut lua_State, tp: c_int) -> *mut c_schar;
+    pub fn lua_typename(state: *mut lua_State, tp: c_int) -> *mut c_char;
 
     /// Pops a table from the stack and sets it as the new metatable for the
     /// value at the given acceptable `index`.
@@ -336,6 +368,14 @@ extern "C" {
     /// - [`LUA_ERRERR`]: error while running the error handler function.
     pub fn lua_pcall(l: *mut lua_State, nargs: c_int, nresults: c_int, errfunc: c_int) -> c_int;
 
+    /// Calls the C function `func` in protected mode. `func` starts with only
+    /// one element in its stack, a light userdata containing `ud`. In case of
+    /// errors, lua_cpcall returns the same error codes as lua_pcall, plus the
+    /// error object on the top of the stack; otherwise, it returns zero, and
+    /// does not change the stack. All values returned by `func` are discarded.
+    /// *[-0, +(0|1), -]*
+    pub fn lua_cpcall(l: *mut lua_State, func: lua_CFunction, ud: *mut c_void) -> c_int;
+
     /// [-0, +1, -]
     /// Loads a Lua chunk. If there are no errors, `lua_load` pushes the
     /// compiled chunk as a Lua function on top of the stack. Otherwise, it
@@ -416,7 +456,7 @@ extern "C" {
 
     // lauxlib functions.
     pub fn luaL_newstate() -> *mut lua_State;
-    pub fn luaL_register(l: *mut lua_State, libname: *const c_schar, lr: *const luaL_Reg);
+    pub fn luaL_register(l: *mut lua_State, libname: *const c_char, lr: *const luaL_Reg);
 
     /// Raises an error. The error message format is given by `fmt` plus any
     /// extra arguments, following the same rules of `lua_pushfstring`. It also
@@ -426,7 +466,7 @@ extern "C" {
     ///
     /// This function never returns, but it is an idiom to use it in C functions
     /// as return `luaL_error(args)`.
-    pub fn luaL_error(l: *mut lua_State, fmt: *const c_schar, ...) -> c_int;
+    pub fn luaL_error(l: *mut lua_State, fmt: *const c_char, ...) -> c_int;
     pub fn luaL_openlibs(L: *mut lua_State);
 
     /// Creates and returns a reference, in the table at index `t`, for the
@@ -456,14 +496,14 @@ extern "C" {
 #[inline(always)]
 /// Pushes onto the stack the value of the global `name`.
 /// *[-0, +1, e]*
-pub unsafe fn lua_getglobal(state: *mut lua_State, name: *const c_schar) {
+pub unsafe fn lua_getglobal(state: *mut lua_State, name: *const c_char) {
     lua_getfield(state, LUA_GLOBALSINDEX, name);
 }
 
 #[inline(always)]
 /// Pops a value from the stack and sets it as the new value of global `name`.
 /// *[-1, +0, e]*
-pub unsafe fn lua_setglobal(state: *mut lua_State, name: *const c_schar) {
+pub unsafe fn lua_setglobal(state: *mut lua_State, name: *const c_char) {
     lua_setfield(state, LUA_GLOBALSINDEX, name);
 }
 
@@ -476,7 +516,7 @@ pub unsafe fn lua_pop(state: *mut lua_State, n: c_int) {
 /// Pushes a C function onto the stack. This function receives a pointer to a C
 /// function and pushes onto the stack a Lua value of type function that, when
 /// called, invokes the corresponding C function.
-/// `[-0, +1, m]`
+/// *[-0, +1, m]*
 ///
 /// Any function to be registered in Lua must follow the correct protocol to
 /// receive its parameters and return its results (see [`lua_CFunction`]).
@@ -485,11 +525,14 @@ pub unsafe fn lua_pushcfunction(state: *mut lua_State, f: lua_CFunction) {
 }
 
 #[inline(always)]
-pub unsafe fn lua_tostring(state: *mut lua_State, i: c_int) -> *const c_schar {
+pub unsafe fn lua_tostring(state: *mut lua_State, i: c_int) -> *const c_char {
     lua_tolstring(state, i, null_mut())
 }
 
 #[inline(always)]
+/// Creates a new empty table and pushes it onto the stack. It is equivalent to
+/// [`lua_createtable`]`(L, 0, 0)`.
+/// *[-0, +1, m]*
 pub unsafe fn lua_newtable(state: *mut lua_State) {
     lua_createtable(state, 0, 0);
 }
@@ -628,7 +671,24 @@ extern "C" {
     /// `ctypename` is a C type name as string (e.g. "struct request",
     /// "uint32_t", etc.).
     /// See also: [`luaL_pushcdata`], [`luaL_checkcdata`]
-    pub fn luaL_ctypeid(l: *mut lua_State, ctypename: *const c_schar) -> u32;
+    pub fn luaL_ctypeid(l: *mut lua_State, ctypename: *const c_char) -> u32;
+
+    /// Check if value at `index` is a lua function, a lua value with a `__call`
+    /// metamethod or a callable `cdata`.
+    pub fn luaL_iscallable(l: *mut lua_State, index: i32) -> i32;
+
+    /// Push the metafield `field` of the value at given `index` onto the stack.
+    /// Returns `1` if the value was pushed, `0` otherwise.
+    /// *[-0, +(1|0), -]*
+    pub fn luaL_getmetafield(l: *mut lua_State, index: i32, field: *const c_char) -> i32;
+}
+
+/// Check if value at given `index` has metafield `field`.
+/// *[-0, +0, -]*
+#[inline(always)]
+#[allow(non_snake_case)]
+pub unsafe fn luaL_hasmetafield(l: *mut lua_State, index: i32, field: *const c_char) -> bool {
+    luaL_getmetafield(l, index, field) != 0 && { lua_pop(l, 1); true }
 }
 
 extern "C" {
@@ -636,6 +696,6 @@ extern "C" {
     /// other measures didn't work and return it. Sets the `len` if it's not
     /// `NULL`. The newly created string is left on top of the stack.
     /// *[-0, +1, m]*
-    pub fn luaT_tolstring(l: *mut lua_State, idx: c_int, len: *mut usize) -> *const c_schar;
+    pub fn luaT_tolstring(l: *mut lua_State, idx: c_int, len: *mut usize) -> *const c_char;
 }
 
