@@ -1,5 +1,3 @@
-use std::error::Error;
-use std::fmt;
 use std::io::Cursor;
 use std::io::Read;
 use std::io::Error as IoError;
@@ -8,18 +6,18 @@ use std::num::NonZeroI32;
 use crate::{
     ffi,
     c_ptr,
-    AbsoluteIndex,
     AsLua,
+    impl_object,
     LuaState,
     LuaRead,
     LuaError,
+    object::{Call, CallError, FromObject, Object},
+    nzi32,
     Push,
     PushInto,
     PushGuard,
     PushOne,
     PushOneInto,
-    ToString,
-    Void,
 };
 
 /// Wrapper around a `&str`. When pushed, the content will be parsed as Lua code and turned into a
@@ -155,7 +153,6 @@ where
             }
 
             let error_msg: String = LuaRead::lua_read(pushed_value.as_lua())
-                .ok()
                 .expect("can't find error message at the top of the Lua stack");
 
             if load_return_value == ffi::LUA_ERRMEM {
@@ -198,31 +195,27 @@ where
 // TODO: example for how to get a LuaFunction as a parameter of a Rust function
 #[derive(Debug)]
 pub struct LuaFunction<L> {
-    lua: L,
-    index: AbsoluteIndex,
+    inner: Object<L>,
 }
 
 impl<L> LuaFunction<L>
 where
     L: AsLua,
 {
-    fn new(lua: L, index: NonZeroI32) -> Self {
-        Self {
-            index: AbsoluteIndex::new(index, lua.as_lua()),
-            lua,
-        }
+    unsafe fn new(lua: L, index: NonZeroI32) -> Self {
+        Self::from_obj(Object::new(lua, index))
     }
 
     pub fn into_inner(self) -> L {
-        self.lua
+        self.inner.into_guard()
     }
 }
 
-impl<L: AsLua> AsLua for LuaFunction<L> {
-    #[inline]
-    fn as_lua(&self) -> LuaState {
-        self.lua.as_lua()
+impl_object!{ LuaFunction,
+    check(lua, index) {
+        ffi::lua_isfunction(lua.as_lua(), index.into())
     }
+    impl Call,
 }
 
 impl<'lua, L> LuaFunction<L>
@@ -243,11 +236,7 @@ where
     where
         V: LuaRead<PushGuard<&'lua L>>,
     {
-        match Self::call_impl(&self.lua, self.index.into(), ()) {
-            Ok(v) => Ok(v),
-            Err(LuaFunctionCallError::LuaError(err)) => Err(err),
-            Err(LuaFunctionCallError::PushError(_)) => unreachable!(),
-        }
+        Call::call(self)
     }
 
     /// Calls the function taking ownership of the underlying push guard.
@@ -265,12 +254,7 @@ where
     where
         V: LuaRead<PushGuard<Self>>,
     {
-        let index = self.index.into();
-        match Self::call_impl(self, index, ()) {
-            Ok(v) => Ok(v),
-            Err(LuaFunctionCallError::LuaError(err)) => Err(err),
-            Err(LuaFunctionCallError::PushError(_)) => unreachable!(),
-        }
+        Call::into_call(self)
     }
 
     /// Calls the function with parameters.
@@ -322,13 +306,12 @@ where
     /// assert_eq!(excess_results, (4, 2, None));
     /// ```
     #[inline]
-    pub fn call_with_args<V, A>(&'lua self, args: A)
-        -> Result<V, LuaFunctionCallError<<A as Push<LuaState>>::Err>>
+    pub fn call_with_args<V, A>(&'lua self, args: A) -> Result<V, CallError<A::Err>>
     where
-        A: Push<LuaState>,
+        A: PushInto<LuaState>,
         V: LuaRead<PushGuard<&'lua L>>,
     {
-        Self::call_impl(&self.lua, self.index.into(), args)
+        Call::call_with(self, args)
     }
 
     /// Calls the function with parameters taking ownership of the underlying
@@ -375,64 +358,19 @@ where
     /// assert_eq!(all_result, (4, 2));
     /// ```
     #[inline]
-    pub fn into_call_with_args<V, A>(self, args: A)
-        -> Result<V, LuaFunctionCallError<<A as Push<LuaState>>::Err>>
+    pub fn into_call_with_args<V, A>(self, args: A) -> Result<V, CallError<A::Err>>
     where
-        A: Push<LuaState>,
+        A: PushInto<LuaState>,
         V: LuaRead<PushGuard<Self>>,
     {
-        let index = self.index.into();
-        Self::call_impl(self, index, args)
-    }
-
-    #[inline]
-    pub fn call_impl<T, V, A, E>(
-        this: T,
-        index: i32,
-        args: A,
-    ) -> Result<V, LuaFunctionCallError<E>>
-    where
-        T: AsLua,
-        A: Push<LuaState, Err = E>,
-        V: LuaRead<PushGuard<T>>,
-    {
-        let raw_lua = this.as_lua();
-        // calling pcall pops the parameters and pushes output
-        let (pcall_return_value, pushed_value) = unsafe {
-            let old_top = ffi::lua_gettop(raw_lua);
-            // lua_pcall pops the function, so we have to make a copy of it
-            ffi::lua_pushvalue(raw_lua, index);
-            let num_pushed = match args.push_to_lua(this.as_lua()) {
-                Ok(g) => g.forget_internal(),
-                Err((err, _)) => return Err(LuaFunctionCallError::PushError(err)),
-            };
-            let pcall_return_value = ffi::lua_pcall(
-                raw_lua,
-                num_pushed,
-                ffi::LUA_MULTRET,
-                0,
-            );
-            let n_results = ffi::lua_gettop(raw_lua) - old_top;
-            (pcall_return_value, PushGuard::new(this, n_results))
-        };
-
-        match pcall_return_value {
-            ffi::LUA_ERRMEM => panic!("lua_pcall returned LUA_ERRMEM"),
-            ffi::LUA_ERRRUN => {
-                let error_msg: ToString = LuaRead::lua_read(pushed_value)
-                    .ok()
-                    .expect("can't find error message at the top of the Lua stack");
-                return Err(LuaError::ExecutionError(error_msg.into()).into())
-            }
-            0 => {}
-            _ => panic!("Unknown error code returned by lua_pcall: {}", pcall_return_value),
-        }
-
-        let n_results = pushed_value.size;
-        LuaRead::lua_read_at_maybe_zero_position(pushed_value, -n_results)
-            .map_err(|lua| LuaError::wrong_type::<V, _>(lua, n_results).into())
+        Call::into_call_with(self, args)
     }
+}
 
+impl<L> LuaFunction<PushGuard<L>>
+where
+    L: AsLua,
+{
     /// Builds a new `LuaFunction` from the code of a reader.
     ///
     /// Returns an error if reading from the `Read` object fails or if there is a syntax error in
@@ -450,11 +388,9 @@ where
     /// assert_eq!(ret, 8);
     /// ```
     #[inline]
-    pub fn load_from_reader(lua: L, code: impl Read)
-        -> Result<LuaFunction<PushGuard<L>>, LuaError>
-    {
+    pub fn load_from_reader(lua: L, code: impl Read) -> Result<Self, LuaError> {
         match LuaCodeFromReader(code).push_into_lua(lua) {
-            Ok(pushed) => Ok(LuaFunction::new(pushed, crate::NEGATIVE_ONE)),
+            Ok(pushed) => unsafe { Ok(Self::new(pushed, nzi32!(-1))) }
             Err((err, _)) => Err(err),
         }
     }
@@ -464,127 +400,9 @@ where
     /// > **Note**: This is just a wrapper around `load_from_reader`. There is no advantage in
     /// > using `load` except that it is more convenient.
     // TODO: remove this function? it's only a thin wrapper and it's for a very niche situation
-    #[inline]
-    pub fn load(lua: L, code: &str) -> Result<LuaFunction<PushGuard<L>>, LuaError> {
+    #[inline(always)]
+    pub fn load(lua: L, code: &str) -> Result<Self, LuaError> {
         let reader = Cursor::new(code.as_bytes());
-        LuaFunction::load_from_reader(lua, reader)
+        Self::load_from_reader(lua, reader)
     }
 }
-
-/// Error that can happen when calling a `LuaFunction`.
-// TODO: implement Error on this
-#[derive(Debug)]
-pub enum LuaFunctionCallError<E> {
-    /// Error while executing the function.
-    LuaError(LuaError),
-    /// Error while pushing one of the parameters.
-    PushError(E),
-}
-
-impl<E> fmt::Display for LuaFunctionCallError<E>
-    where E: fmt::Display
-{
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match *self {
-            LuaFunctionCallError::LuaError(ref lua_error) => write!(f, "Lua error: {}", lua_error),
-            LuaFunctionCallError::PushError(ref err) => {
-                write!(f, "Error while pushing arguments: {}", err)
-            }
-        }
-    }
-}
-
-impl<E> From<LuaError> for LuaFunctionCallError<E> {
-    #[inline]
-    fn from(err: LuaError) -> LuaFunctionCallError<E> {
-        LuaFunctionCallError::LuaError(err)
-    }
-}
-
-impl<E> From<LuaFunctionCallError<E>> for LuaError
-where
-    E: Into<Void>,
-{
-    fn from(err: LuaFunctionCallError<E>) -> LuaError {
-        match err {
-            LuaFunctionCallError::LuaError(lua_error) => lua_error,
-            LuaFunctionCallError::PushError(_) => unreachable!("Void cannot be instantiated"),
-        }
-    }
-}
-
-impl<E> Error for LuaFunctionCallError<E>
-    where E: Error
-{
-    fn description(&self) -> &str {
-        match *self {
-            LuaFunctionCallError::LuaError(_) => "Lua error",
-            LuaFunctionCallError::PushError(_) => "error while pushing arguments",
-        }
-    }
-
-    fn cause(&self) -> Option<&dyn Error> {
-        match *self {
-            LuaFunctionCallError::LuaError(ref lua_error) => Some(lua_error),
-            LuaFunctionCallError::PushError(ref err) => Some(err),
-        }
-    }
-}
-
-impl Error for LuaFunctionCallError<Void> {
-    fn description(&self) -> &str {
-        match *self {
-            LuaFunctionCallError::LuaError(_) => "Lua error",
-            _ => unreachable!("Void cannot be instantiated"),
-        }
-    }
-
-    fn cause(&self) -> Option<&dyn Error> {
-        match *self {
-            LuaFunctionCallError::LuaError(ref lua_error) => Some(lua_error),
-            _ => unreachable!("Void cannot be instantiated"),
-        }
-    }
-}
-
-// TODO: return Result<Ret, ExecutionError> instead
-// impl<'a, 'lua, Ret: CopyRead> ::std::ops::FnMut<(), Ret> for LuaFunction<'a,'lua> {
-// fn call_mut(&mut self, _: ()) -> Ret {
-// self.call().unwrap()
-// }
-// }
-
-impl<L> LuaRead<L> for LuaFunction<L>
-where
-    L: AsLua,
-{
-    #[inline]
-    fn lua_read_at_position(lua: L, index: NonZeroI32) -> Result<LuaFunction<L>, L> {
-        if unsafe { ffi::lua_isfunction(lua.as_lua(), index.get()) } {
-            Ok(LuaFunction::new(lua, index))
-        } else {
-            Err(lua)
-        }
-    }
-}
-
-impl<L, T> Push<L> for LuaFunction<T>
-where
-    L: AsLua,
-{
-    type Err = Void;
-
-    fn push_to_lua(&self, lua: L) -> crate::PushResult<L, Self> {
-        unsafe {
-            ffi::lua_pushvalue(lua.as_lua(), self.index.into());
-            Ok(PushGuard::new(lua, 1))
-        }
-    }
-}
-
-impl<L, T> PushOne<L> for LuaFunction<T>
-where
-    L: AsLua,
-{
-}
-
